#include <stdio.h>
#include "parti.h"

#undef MAX
#undef MIN
#define MAX(a,b)  ((a)>(b))?(a):(b) 
#define MIN(a,b)  ((a)<(b))?(a):(b) 

void 
new_execute_gather(pattern, buf_list, data, size)
struct com_pattern *pattern;
char **buf_list,*data;
int size;
{
  int  node;				/* processor id of incoming message */
  int  *msg_id;				/* message id of irecv */ 
  int  i,j,p;
  int  count;
  int  *l;
  char sync_buf[10];			/* buffer for synchronizing messages */
  char *buffer, *ptr;
  int    *ibuf, *idata;			/* pointers for integer */
  double *dbuf, *ddata;			/* pointers for double */
   
#ifdef DEBUG
  trace_enter("new_execute_gather");
#endif

  /* message buffer */
  msg_id = wbuffer->msg_id;
  if (size > sizeof(double))
    buffer = (char *) malloc(pattern->buffer_size*size);
  else
    buffer = (char *) wbuffer->buffer;

  /* check if processor send messages to itself */
  if (pattern->send_n[PARTI_mynode]) {
    /* copy data to buffer */
    l = pattern->send_list[PARTI_mynode];
    switch(size) {
    case sizeof(int):
      ibuf  = (int *) buf_list[PARTI_mynode];
      idata = (int *) data;
      for (j=0; j<pattern->send_n[PARTI_mynode]; j++)
        ibuf[j] = idata[l[j]];
      break;
    case sizeof(double):
      dbuf  = (double *) buf_list[PARTI_mynode];
      ddata = (double *) data;
      for (j=0; j<pattern->send_n[PARTI_mynode]; j++)
        dbuf[j] = ddata[l[j]];
      break;
    case sizeof(char):
      ptr = buf_list[PARTI_mynode];
      for (j=0; j<pattern->send_n[PARTI_mynode]; j++)
        ptr[j] = data[l[j]];
      break;
    default:
      ptr = buf_list[PARTI_mynode];
      for (j=0; j<pattern->send_n[PARTI_mynode]; j++,ptr+=size)
        bytecopy(ptr, &data[l[j]*size], size);
      break;
    }
  }
    
  /* call asynchronous receive funcations */
  for (i=1, count=0; i<PARTI_numnodes; i++){
    p = pattern->gcolor[i];
    if (pattern->fetch_n[p]) {
      msg_id[count++] = PARTI_irecv (FORCED_GATHER_EXECUTOR_MSG+p,
	           buf_list[p], pattern->fetch_n[p]*size, p, PARTI_pid);
    }

      PARTI_csend (GATHER_SYNC+PARTI_mynode, 
		   sync_buf, 1, p, PARTI_pid);

      PARTI_crecv (GATHER_SYNC+p, sync_buf, 1, p, PARTI_pid);

    /* send data to other processors */
    if(pattern->send_n[p]){
      l = pattern->send_list[p];
      switch(size) {
      case sizeof(int):
        ibuf  = (int *) buffer;
        idata = (int *) data;
        for (j=0; j<pattern->send_n[p]; j++)
          ibuf[j] = idata[l[j]];
        break;
      case sizeof(double):
        dbuf  = (double *) buffer;
        ddata = (double *) data;
        for (j=0; j<pattern->send_n[p]; j++)
          dbuf[j] = ddata[l[j]];
        break;
      case sizeof(char):
        for (j=0; j<pattern->send_n[p]; j++)
          buffer[j] = data[l[j]];
        break;
      default:
        for(j=0,ptr=buffer;j<pattern->send_n[p];j++,ptr+=size)
          bytecopy(ptr,&data[l[j]*size],size);
        break;
      }
      PARTI_bsend (FORCED_GATHER_EXECUTOR_MSG+PARTI_mynode,
	     buffer, size*pattern->send_n[p], p, PARTI_pid);
    }  
  }

  /* wait until all messages are done */
  for (i = 0; i < count; i++) {
    PARTI_msgwait (msg_id[i]);
  }

  /* free buffer is necessary */
  if (size > sizeof(double))
    free (buffer);

#ifdef DEBUG
  trace_exit();
#endif
}

void 
new_execute_scatter(pattern, buf_list, data, size, func)
struct com_pattern *pattern;
char **buf_list,*data;
int size;
void (*func)();
{
  int  node;				/* processor id of incoming message */
  int  *msg_id;				/* message id of isend */
  int  i,j,p;
  int  count;
  int  *l;
  char *buffer, *ptr;
  char **wlist;				/* pointer of buffer */
  char sync_buf[1];			/* buffer for synchronizing messages */
  void scatter_func();
  double *dbuf, *ddata;
  double start, end;

#ifdef DEBUG
  trace_enter("new_execute_scatter");
#endif

  /* message buffer */
  msg_id = wbuffer->msg_id;
  if (size > sizeof(double))
    buffer = (char *) malloc(pattern->total_send_size*size);
  else
    buffer = (char *) wbuffer->buffer;

  /* check if processor send messages to itself */
  if (pattern->fetch_n[PARTI_mynode]) {
    /* copy the buffer to target */
    scatter_func(data, buf_list[PARTI_mynode], pattern->send_list[PARTI_mynode],
		 pattern->send_n[PARTI_mynode], size, func);
  }

  /* set up the buffer list */
  wlist = wbuffer->wlist;
  for (p = 0, count = 0; p < PARTI_numnodes; p++) {
    wlist[p] = &buffer[count];
    if (p != PARTI_mynode)
      count += (pattern->send_n[p]*size);
  }
    
  /* send and receive messages */
  for (i=1, count=0; i<PARTI_numnodes; i++){
    p = pattern->gcolor[i];
    /* call asynchronous receive funcations */
    if (pattern->send_n[p]) {
      msg_id[count++] =
      PARTI_irecv (FORCED_SCATTER_EXECUTOR_MSG+p,
	           wlist[p], pattern->send_n[p]*size, p, PARTI_pid);
      PARTI_csend (SCATTER_SYNC+PARTI_mynode, 
		   sync_buf, 0, p, PARTI_pid);
    }
    
    /* send data */
    if (pattern->fetch_n[p]) {
      PARTI_crecv (SCATTER_SYNC+p, sync_buf, 0, p, PARTI_pid);
      PARTI_bsend (FORCED_SCATTER_EXECUTOR_MSG+PARTI_mynode,
		   buf_list[p], size*pattern->fetch_n[p], p, PARTI_pid);
    }
  }

  /* wait messages and store data */
  for (i = 1, count = 0; i < PARTI_numnodes; i++) {
    p = pattern->gcolor[i];
    if (pattern->send_n[p]) {
      PARTI_msgwait(msg_id[count]);
      scatter_func(data, wlist[p], pattern->send_list[p], 
		 pattern->send_n[p], size, func);
      count++;
    }
  }

  /* free the buffer is necessary */
  if (size > sizeof(double))
    free (buffer);

#ifdef DEBUG 
  trace_exit(); 
#endif    
}

void
scatter_func(data, buffer, index, ndata, size, func)
char  *data, *buffer;
int   *index;
int   ndata, size;
void  (*func)();

{
  int    i;
  int    *is, *it;		/* integer source and target arrays */
  float  *fs, *ft;		/* integer source and target arrays */
  double *ds, *dt;		/* integer source and target arrays */

  switch ((int) func) {
  case NULL:			/* no operations */
    switch (size) {
    case sizeof(int):
      for (i = 0, is = (int *) buffer, it = (int *) data; i < ndata; i++)
	it[index[i]] = is[i];
      break;
    case sizeof(double):
      for (i = 0, ds = (double *) buffer, dt = (double *) data; i < ndata; i++)
	dt[index[i]] = ds[i];
    case sizeof(char):
      for (i = 0; i < ndata; i++)
	data[index[i]] = buffer[i];
    default:
      for (i = 0; i < ndata; i++, buffer+=size)
        bytecopy(&data[index[i]*size], buffer, size);
      break;
    }
    break;
  
  case PARTI_add:		/* integer addition */
    for (i = 0, is = (int *) buffer, it = (int *) data; i < ndata; i++)
      it[index[i]] += is[i];
    break;
  
  case PARTI_sub:		/* integer substraction */
    for (i = 0, is = (int *) buffer, it = (int *) data; i < ndata; i++)
      it[index[i]] -= is[i];
    break;
  
  case PARTI_mul:		/* integer multiplication */
    for (i = 0, is = (int *) buffer, it = (int *) data; i < ndata; i++)
      it[index[i]] *= is[i];
    break;

  case PARTI_max:               /* integer max */
    for (i = 0, is = (int *) buffer, it = (int *) data; i < ndata; i++)
      it[index[i]] = MAX(it[index[i]],is[i]);
    break;

  case PARTI_min:               /* integer max */
    for (i = 0, is = (int *) buffer, it = (int *) data; i < ndata; i++)
      it[index[i]] = MIN(it[index[i]],is[i]);
    break;

  case PARTI_logior:               /* integer logical inclusive or */
    for (i = 0, is = (int *) buffer, it = (int *) data; i < ndata; i++)
      it[index[i]] |= is[i];
    break;

  case PARTI_logand:               /* integer logical and */
    for (i = 0, is = (int *) buffer, it = (int *) data; i < ndata; i++)
      it[index[i]] &= is[i];
    break;

  case PARTI_logxor:               /* integer logical exclusive or */
    for (i = 0, is = (int *) buffer, it = (int *) data; i < ndata; i++)
      it[index[i]] ^= is[i];
    break;

  
  case PARTI_cadd:		/* character addition */
    for (i = 0; i < ndata; i++)
      data[index[i]] += buffer[i];
    break;
  
  case PARTI_csub:		/* character substraction */
    for (i = 0; i < ndata; i++)
      data[index[i]] -= buffer[i];
    break;
  
  case PARTI_cmul:		/* character multiplication */
    for (i = 0; i < ndata; i++)
      data[index[i]] *= buffer[i];
    break;

  case PARTI_cmax:              /* character multiplication */
    for (i = 0; i < ndata; i++)
      data[index[i]] = MAX(data[index[i]],buffer[i]);
    break;

  case PARTI_cmin:              /* character multiplication */
    for (i = 0; i < ndata; i++)
      data[index[i]] = MIN(data[index[i]],buffer[i]);
    break;

  case PARTI_clogior:              /* character logical incl or */
    for (i = 0; i < ndata; i++)
      data[index[i]] |= buffer[i];
    break;

  case PARTI_clogand:              /* character logical add */
    for (i = 0; i < ndata; i++)
      data[index[i]] &= buffer[i];
    break;

  case PARTI_clogxor:              /* character logical excl or */
    for (i = 0; i < ndata; i++)
      data[index[i]] ^= buffer[i];
    break;

  
  case PARTI_fadd:		/* float addition */
    for (i = 0, fs = (float *) buffer, ft = (float *) data; i < ndata; i++)
      ft[index[i]] += fs[i];
    break;
  
  case PARTI_fsub:		/* float substraction */
    for (i = 0, fs = (float *) buffer, ft = (float *) data; i < ndata; i++)
      ft[index[i]] -= fs[i];
    break;
  
  case PARTI_fmul:		/* float multiplication */
    for (i = 0, fs = (float *) buffer, ft = (float *) data; i < ndata; i++)
      ft[index[i]] *= fs[i];
    break;

  case PARTI_fmax:              /* float multiplication */
    for (i = 0, fs = (float *) buffer, ft = (float *) data; i < ndata; i++)
      ft[index[i]] = MAX(ft[index[i]], fs[i]);
    break;

  case PARTI_fmin:              /* float multiplication */
    for (i = 0, fs = (float *) buffer, ft = (float *) data; i < ndata; i++)
      ft[index[i]] = MIN(ft[index[i]], fs[i]);
    break;

  case PARTI_flogior:		/* float logical incl or */
	printf("Error : No IOR for floats\n");
    break;

  case PARTI_flogand:		/* float logical and */
	printf("Error : No AND for floats\n");
    break;

  case PARTI_flogxor:		/* float logical excl or */
	printf("Error : No XOR for floats\n");
    break;

  
  case PARTI_dadd:		/* double addition */
    for (i = 0, ds = (double *) buffer, dt = (double *) data; i < ndata; i++)
      dt[index[i]] += ds[i];
    break;
  
  case PARTI_dsub:		/* double substraction */
    for (i = 0, ds = (double *) buffer, dt = (double *) data; i < ndata; i++)
      dt[index[i]] -= ds[i];
    break;
  
  case PARTI_dmul:		/* double multiplication */
    for (i = 0, ds = (double *) buffer, dt = (double *) data; i < ndata; i++)
      dt[index[i]] *= ds[i];
    break;

  case PARTI_dmax:		/* double multiplication */
    for (i = 0, ds = (double *) buffer, dt = (double *) data; i < ndata; i++)
      dt[index[i]] = MAX(dt[index[i]],ds[i]);
    break;

  case PARTI_dmin:		/* double multiplication */
    for (i = 0, ds = (double *) buffer, dt = (double *) data; i < ndata; i++)
      dt[index[i]] = MIN(dt[index[i]],ds[i]);
    break;

  case PARTI_dlogior:		/* double logical incl or */
	printf("Error : No IOR for double\n");
    break;

  case PARTI_dlogand:		/* double logical and */
	printf("Error : No AND for double\n");
    break;

  case PARTI_dlogxor:		/* double logical excl or */
	printf("Error : No XOR for double\n");
    break;


  default:			/* user defined operations */
    for (i = 0; i < ndata; i++, buffer+=size)
      (*func)(&data[index[i]*size], buffer);
    break;
  }
}
